# Load the benchmark data set into the database
import sys
import gzip
import MySQLdb

BENCHMARK_DATABASE = "cyclops_mysql_benchmark"

MYSQL_CONNECT_ARGS = dict(
    host = "localhost",
    user = "root",
    passwd = "",
    )

def C(database=None):
    args = MYSQL_CONNECT_ARGS.copy()
    if database is not None:
        args["db"] = database
    db = MySQLdb.connect(**args)
    return db.cursor()


def create_benchmark_database():
    c = C()
    c.execute("create database " + BENCHMARK_DATABASE);
    c = C(BENCHMARK_DATABASE)
    c.execute("""
CREATE TABLE compound (
  idx INTEGER PRIMARY KEY,
  smiles TEXT,
  title TEXT,
  maccs TEXT,
  fp TEXT)
""")

def load_benchmark_database():
    c = C(BENCHMARK_DATABASE)
    c.execute("select count(*) from compound")
    start = c.fetchone()[0]
    if start != 0:
        print "Warning:", start, "structures already in the database table"
        
    f = gzip.open("nci_09425001_09450000.smi.gz")
    data = []
    idx = start
    for line in f:
        fields = line.split()
        data.append( dict(idx=idx, smiles=fields[0], title=fields[1]) )
        idx += 1

    c.executemany(
"INSERT INTO compound (idx, smiles, title, maccs, fp) "
" VALUE (%(idx)s, %(smiles)s, %(title)s, oe_maccs_fp(%(smiles)s), oe_path_fp(%(smiles)s))",
                  data)
    c.execute("select count(*) from compound")
    end = c.fetchone()[0]

    num_loaded = end-start
    print "Loaded", num_loaded, "structures."
    if start != 0:
        print end, "structures in the database"

def size_benchmark_database():
    c = C(BENCHMARK_DATABASE)
    c.execute("select count(*) from compound")
    size = c.fetchone()[0]
    print "The benchmark database contains", size, "structures."

def clear_benchmark_database():
    c = C(BENCHMARK_DATABASE)
    c.execute("delete from compound")


def drop_benchmark_database():
    c = C()
    c.execute("drop database " + BENCHMARK_DATABASE)


import optparse
parser = optparse.OptionParser()
parser.add_option("--create", dest="create", action="store_true",
                  help = "create the benchmark database")
parser.add_option("--load", dest="load", action="store_true",
                  help = "load all the structures into the database")
parser.add_option("--size", dest="size", action="store_true",
                  help = "give the number of structures in the database")
parser.add_option("--clear", dest="clear", action="store_true",
                  help = "remove all structures from the database")
parser.add_option("--drop", dest="drop", action="store_true",
                  help = "remove the benchmark database")



def main():
    if len(sys.argv) == 1:
        create_benchmark_database()
        load_benchmark_database()
        return
        
    (options, args) = parser.parse_args()
    if args:
        parser.error("no extra arguments are allowed")
    if options.create:
        create_benchmark_database()
    if options.load:
        load_benchmark_database()
    if options.size:
        size_benchmark_database()
    if options.clear:
        clear_benchmark_database()
    if options.drop:
        drop_benchmark_database()

if __name__ == "__main__":
    main()
