import unittest
import MySQLdb
import re
import datetime
import traceback

# These need to
MYSQL_CONNECT_ARGS = dict(
    host = "localhost",
    user = "root",
    passwd = "",
    )

TEST_DATABASE = "cyclops_mysql_testdb"

def C(database=None):
    args = MYSQL_CONNECT_ARGS.copy()
    if database is not None:
        args["db"] = database
    db = MySQLdb.connect(**args)
    return db.cursor()

# "E"xec a SQL statement. Must return a single row. Return that row.
def E(sql, *args):
    conn = C()
    conn.execute(sql, *args)
    result = conn.fetchall()
    assert len(result) == 1
    return result[0]

# Connect to a live database. Returns a "L"ist, hence the "L"
def L(sql, *args):
    conn = C(TEST_DATABASE)
    #print
    #print sql
    #print
    conn.execute(sql, *args)
    return conn.fetchall()

def assertErrorMessage(self, sql, msg_contains):
    try:
        E(sql)
    except MySQLdb.Error, ex:
        msg = ex[1]
        self.assert_(msg_contains in msg, "Cannot find %r in %r" % (msg_contains, msg))
        return
    self.fail("Expected exception from %r" % (sql,))

date_pat = re.compile(r"(?P<year>\d{4})-(?P<month>\d\d)-(?P<day>\d\d)$")

class TestOELicensed(unittest.TestCase):
    def test_oe_licensed(self):
        # If these don't work then other tests will fail
        self.assertEquals(E("select oe_licensed()"), (1,))
        self.assertEquals(E("select oe_licensed('cyclops')"), (1,))
        self.assertEquals(E("select oe_licensed('oechem')"), (1,))
        self.assertEquals(E("select oe_licensed('oegraphsim')"), (1,))
        self.assertEquals(E("select oe_licensed('oegraphsim')"), (1,))
        self.assertEquals(E("select oe_licensed(concat(char(111),'egraphsim'))"), (1,))
    def test_oe_licensed_bad_args(self):
        assertErrorMessage(self, "select oe_licensed('oechem', 'oechem')",
                           "too many arguments")
        assertErrorMessage(self, "select oe_licensed(1)", "must be a string")
        assertErrorMessage(self, "select oe_licensed('oe')", "unknown module")
        assertErrorMessage(self, "select oe_licensed('')", "unknown module")
        assertErrorMessage(self, "select oe_licensed(concat(char(111), 'egraphsi'))",
                           "unknown module name")

def _check_date(self, s):
    # is the date in the right format?
    m = date_pat.match(s)
    self.assert_(m is not None, "did not match pattern")
    # Check that the date is valid
    y, m, d = map(int, m.groups())
    datetime.datetime(y, m, d)
    self.assert_(2000 < y < 3000, "bad year")
        
class TestOELicenseDate(unittest.TestCase):
    def test_oe_license_date(self):
        result = E("select oe_license_date('oechem')")
        self.assertEquals(len(result), 1)
        _check_date(self, result[0])
        
        result = E("select oe_license_date('oegraphsim')")
        _check_date(self, result[0])
        
        result = E("select oe_license_date('cyclops')")
        self.assertEquals(result[0], "9999-12-31")


    def test_oe_license_date_bad_args(self):
        assertErrorMessage(self, "select oe_license_date()", "missing required module")
        assertErrorMessage(self, "select oe_license_date('')", "unknown module")
        assertErrorMessage(self, "select oe_license_date('oegraphsi')", "unknown module")
        assertErrorMessage(self, "select oe_license_date(1)", "must be a string")
        assertErrorMessage(self, "select oe_license_date('cyclops', 1)", "too many")

class TestOEVersion(unittest.TestCase):
    def test_oe_version(self):
        self.assertEquals(E("select oe_version('cyclops')"), ("1.0",))
        s, = E("select oe_version('oechem')")
        _check_date(self, s.split()[0])
        self.assert_(len(s.split()) >= 2, "missing terms in oechem")
        s, = E("select oe_version('oegraphsim')")
        _check_date(self, s.split()[0])
        self.assert_(len(s.split()) >= 2, "missing terms in oegraphsim")

    def test_oe_version_bad_args(self):
        assertErrorMessage(self, "select oe_version()", "missing required module")
        assertErrorMessage(self, "select oe_version('')", "unknown module")
        assertErrorMessage(self, "select oe_version('CYCLOPS')", "unknown module")
        assertErrorMessage(self, "select oe_version(1)", "must be a string")
        assertErrorMessage(self, "select oe_version('cyclops', 1)", "too many")

class TestValidSmiles(unittest.TestCase):
    def test_valid_smiles(self):
        self.assertEquals(E("select oe_valid_smiles('C')"), (1,))
        self.assertEquals(E("select oe_valid_smiles('CC')"), (1,))
        self.assertEquals(E("select oe_valid_smiles('c1ccccc1O')"), (1,))
        # SMILES stops at the first whitespace
        self.assertEquals(E("select oe_valid_smiles('C title')"), (1,))

    def test_invalid_smiles(self):
        self.assertEquals(E("select oe_valid_smiles('C(')"), (0,))
        self.assertEquals(E("select oe_valid_smiles('CQ')"), (0,))
        self.assertEquals(E("select oe_valid_smiles('c1ccccc1O=')"), (0,))
        # In OEChem's strict mode, the empty string is not valid
        self.assertEquals(E("select oe_valid_smiles('')"), (0, ))
        # Neither is something which starts with a space
        self.assertEquals(E("select oe_valid_smiles(' C')"), (0,))


class TestCanSmiles(unittest.TestCase):
    def test_cansmiles1(self):
        x1 = E("select oe_cansmiles('CCO')")
        x2 = E("select oe_cansmiles('OCC')")
        self.assertEquals(x1, x2)
        self.assertEquals(x1[0], "CCO")

    def test_cansmiles2(self):
        x1 = E("select oe_cansmiles('[13CH3]CO')")
        x2 = E("select oe_cansmiles('OCC')")
        self.assertEquals(x1, x2)
        self.assertEquals(x1[0], "CCO")

    def test_cansmiles3(self):
        self.assertEquals(E("select oe_cansmiles('C1=CC=CC=C1')"), ("c1ccccc1",))

    def test_bad_args(self):
        self.assertRaises(MySQLdb.Error, E, "select oe_cansmiles()")
        self.assertRaises(MySQLdb.Error, E, "select oe_cansmiles(1)")
        self.assertRaises(MySQLdb.Error, E, "select oe_cansmiles('C', 'C')")
        self.assertRaises(MySQLdb.Error, E, "select oe_cansmiles('Q')")

class TestIsoSmiles(unittest.TestCase):
    def test_isosmiles1(self):
        x1 = E("select oe_isosmiles('CCO')")
        x2 = E("select oe_isosmiles('OCC')")
        self.assertEquals(x1, x2)
        self.assertEquals(x1[0], "CCO")

    def test_isosmiles2(self):
        x1 = E("select oe_isosmiles('[13CH3]CO')")
        x2 = E("select oe_isosmiles('OCC')")
        self.assertNotEquals(x1, x2)
        self.assertEquals(x1[0], "[13CH3]CO")

    def test_isosmiles3(self):
        self.assertEquals(E("select oe_isosmiles('C1=CC=CC=C1')"), ("c1ccccc1",))

    def test_bad_args(self):
        self.assertRaises(MySQLdb.Error, E, "select oe_isosmiles()")
        self.assertRaises(MySQLdb.Error, E, "select oe_isosmiles(1)")
        self.assertRaises(MySQLdb.Error, E, "select oe_isosmiles('C', 'C')")
        self.assertRaises(MySQLdb.Error, E, "select oe_isosmiles('Q')")

class TestKekSmiles(unittest.TestCase):
    def test_keksmiles1(self):
        x1 = E("select oe_keksmiles('CCO')")
        x2 = E("select oe_keksmiles('OCC')")
        self.assertEquals(x1, x2)
        self.assertEquals(x1[0], "CCO")

    def test_keksmiles2(self):
        x1 = E("select oe_isosmiles('[13CH3]CO')")
        x2 = E("select oe_isosmiles('OCC')")
        self.assertNotEquals(x1, x2)
        self.assertEquals(x1[0], "[13CH3]CO")

    def test_keksmiles3(self):
        # As you can see, the Kekule SMILES are NOT canonical
        self.assertEquals(E("select oe_keksmiles('C1=CC=CC=C1')"), ("C=1C=CC=CC1",))
        self.assertEquals(E("select oe_keksmiles('C=1C=CC=CC=1')"), ("C1=CC=CC=C1",))
        self.assertEquals(E("select oe_keksmiles('c1ccccc1')"), ("C1=CC=CC=C1",))

    def test_bad_args(self):
        self.assertRaises(MySQLdb.Error, E, "select oe_keksmiles()")
        self.assertRaises(MySQLdb.Error, E, "select oe_keksmiles(1)")
        self.assertRaises(MySQLdb.Error, E, "select oe_keksmiles('C', 'C')")
        self.assertRaises(MySQLdb.Error, E, "select oe_keksmiles('Q')")

class TestOEMatchs(unittest.TestCase):
    def test_oematches(self):
        self.assertEquals(E("select oe_matches('C', 'C')"), (1,))
        self.assertEquals(E("select oe_matches('C', 'CC')"), (0,))
        self.assertEquals(E("select oe_matches('CC', 'C')"), (1,))
        self.assertEquals(E("select oe_matches('CC', 'CC')"), (1,))

    def test_bad_args(self):
        self.assertRaises(MySQLdb.Error, E, "select oe_matches()")
        self.assertRaises(MySQLdb.Error, E, "select oe_matches(1)")
        self.assertRaises(MySQLdb.Error, E, "select oe_matches(1, 2)")
        self.assertRaises(MySQLdb.Error, E, "select oe_matches(1, 2, 3)")
        self.assertRaises(MySQLdb.Error, E, "select oe_matches('C', 2)")
        self.assertRaises(MySQLdb.Error, E, "select oe_matches(1, 'C')")
        self.assertRaises(MySQLdb.Error, E, "select oe_matches(1, 'C')")
        self.assertRaises(MySQLdb.Error, E, "select oe_matches('C', 'C', 'C')")

class TestCountMatches(unittest.TestCase):
    def test_count_matches(self):
        self.assertEquals(E("select oe_count_matches('C', 'C')"), (1,))
        self.assertEquals(E("select oe_count_matches('CC', 'C')"), (2,))
        self.assertEquals(E("select oe_count_matches('CCC', 'C')"), (3,))
        self.assertEquals(E("select oe_count_matches('CCC', 'CC')"), (4,))

    def test_bad_args(self):
        self.assertRaises(MySQLdb.Error, E, "select oe_count_matches()")
        self.assertRaises(MySQLdb.Error, E, "select oe_count_matches(1)")
        self.assertRaises(MySQLdb.Error, E, "select oe_count_matches(1, 2)")
        self.assertRaises(MySQLdb.Error, E, "select oe_count_matches(1, 2, 3)")
        self.assertRaises(MySQLdb.Error, E, "select oe_count_matches('C', 2)")
        self.assertRaises(MySQLdb.Error, E, "select oe_count_matches(1, 'C')")
        self.assertRaises(MySQLdb.Error, E, "select oe_count_matches(1, 'C')")
        self.assertRaises(MySQLdb.Error, E, "select oe_count_matches('C', 'C', 'C')")
        

class TestCountUMatches(unittest.TestCase):
    def test_count_umatches(self):
        self.assertEquals(E("select oe_count_umatches('C', 'C')"), (1,))
        self.assertEquals(E("select oe_count_umatches('CC', 'C')"), (2,))
        self.assertEquals(E("select oe_count_umatches('CCC', 'C')"), (3,))
        self.assertEquals(E("select oe_count_umatches('CCC', 'CC')"), (2,))

    def test_bad_args(self):
        self.assertRaises(MySQLdb.Error, E, "select oe_count_umatches()")
        self.assertRaises(MySQLdb.Error, E, "select oe_count_umatches(1)")
        self.assertRaises(MySQLdb.Error, E, "select oe_count_umatches(1, 2)")
        self.assertRaises(MySQLdb.Error, E, "select oe_count_umatches(1, 2, 3)")
        self.assertRaises(MySQLdb.Error, E, "select oe_count_umatches('C', 2)")
        self.assertRaises(MySQLdb.Error, E, "select oe_count_umatches(1, 'C')")
        self.assertRaises(MySQLdb.Error, E, "select oe_count_umatches(1, 'C')")
        self.assertRaises(MySQLdb.Error, E, "select oe_count_umatches('C', 'C', 'C')")

class TestLingoSim(unittest.TestCase):
    def test_lingosim(self):
        self.assertEquals(E("select oe_lingosim('c1ccccc1', 'c1ccccc1O')"),
                          (NearFloat(0.83333331346511841),))
        self.assertEquals(E("select oe_lingosim('c1ccccc1O', 'c1ccccc1')"),
                          (NearFloat(0.83333331346511841),))
    def test_bad_args(self):
        self.assertRaises(MySQLdb.Error, E, "select oe_lingosim()")
        self.assertRaises(MySQLdb.Error, E, "select oe_lingosim('C')")
        self.assertRaises(MySQLdb.Error, E, "select oe_lingosim('C', 'C', 'C')")
        self.assertRaises(MySQLdb.Error, E, "select oe_lingosim(1, 'C')")
        self.assertRaises(MySQLdb.Error, E, "select oe_lingosim('C', 2)")
        self.assertRaises(MySQLdb.Error, E, "select oe_lingosim(1, 2)")


class TestMACCS(unittest.TestCase):
    def test_oe_maccs_fp(self):
        self.assertEquals(E("select oe_maccs_fp('Nc1ccccc1O')"),
                          ("0000000000000000000084040000102405488283b3",))
        # position 4 (bit 3) contains Np. That's the leading 8.
        # I can't explain position 166 (bit 165).
        self.assertEquals(E("select oe_maccs_fp('[Np]')"),
                          ("800000000000000000000000000000000000000002",))
    def test_bad_smiles(self):
        assertErrorMessage(self, "select oe_maccs_fp('1')", "cannot parse SMILES string")
        assertErrorMessage(self, "select oe_maccs_fp('')", "cannot parse SMILES string")
        assertErrorMessage(self, "select oe_maccs_fp(' C')", "cannot parse SMILES string")

    def test_bad_args(self):
        self.assertRaises(MySQLdb.Error, E, "select oe_maccs_fp()")
        self.assertRaises(MySQLdb.Error, E, "select oe_maccs_fp(1)")
        self.assertRaises(MySQLdb.Error, E, "select oe_maccs_fp(9.8)")
        self.assertRaises(MySQLdb.Error, E, "select oe_maccs_fp('O', 'O')")

# Run-length encoding; the character followed by the number of times it repeats
def rle_decode(*args):
    assert len(args) % 2 == 0, args
    s = ""
    for i in range(0, len(args), 2):
        s += args[i]*args[i+1]
    return s

# Helper function to make the rl-encoded string
def rle_encode(s):
    fields = []
    while s:
        c = s[0]
        fields.append(c)
        t = s.lstrip(c)
        fields.append(len(s) - len(t))
        s = t
    return fields

class TestPathFP(unittest.TestCase):
    def test_simple_path(self):
        self.assertEquals(E("select oe_path_fp('C')"),
                          (rle_decode('0', 939, '2', 1, '0', 84),))
        self.assertEquals(E("select oe_path_fp('C', 4096)"),
                          (rle_decode('0', 939, '2', 1, '0', 84),))
        self.assertEquals(E("select oe_path_fp('C', 2048)"),
                          (rle_decode('0', 469, '1', 1, '0', 42),))
        self.assertEquals(E("select oe_path_fp('CC', 1024)"),
                          (rle_decode('0', 141, '1', 1, '0', 92, '1', 1, '0', 21),))
        self.assertEquals(E("select oe_path_fp('C', 16)"), ("0020",))
        self.assertEquals(E("select oe_path_fp('CCCNO', 16)"), ("fe71",))
        self.assertEquals(E("select oe_path_fp('C', 65536)"),
                          (rle_decode('0', 6460, '4', 1, '0', 9923),))

    def test_bad_path_args(self):
        assertErrorMessage(self, "select oe_path_fp()", "not enough arguments")
        assertErrorMessage(self, "select oe_path_fp('C', 1024, 1, 2, 0, 0, 8)", "too many arguments")
        assertErrorMessage(self, "select oe_path_fp(1)",
                           "argument 1 (smiles) must be a SMILES string")
        assertErrorMessage(self, "select oe_path_fp('C', 'X')",
                           "argument 2 (num_bits) must be an integer")
        assertErrorMessage(self, "select oe_path_fp('C', -1024)", "must be between")
        assertErrorMessage(self, "select oe_path_fp('C', 8)", "must be between")
        assertErrorMessage(self, "select oe_path_fp('C', 15)", "must be between")
        assertErrorMessage(self, "select oe_path_fp('C', 65537)", "must be between")
        assertErrorMessage(self, "select oe_path_fp('C', 16, 5, 4)",
             "argument 4 (max_bonds) must not be smaller than argument 3 (min_bonds)")
        
        # check for overflow around 2**31 and 2**32
        assertErrorMessage(self, "select oe_path_fp('C', 2147483647)", "must be between")
        assertErrorMessage(self, "select oe_path_fp('C', 2147483847)", "must be between")
        assertErrorMessage(self, "select oe_path_fp('C', 4294967296)", "must be between")
        assertErrorMessage(self, "select oe_path_fp('C', 4294967400)", "must be between")
        assertErrorMessage(self, "select oe_path_fp('C', 17)", "must be a multiple of 8")
        assertErrorMessage(self, "select oe_path_fp('C', 128, 'C', 5)",
                           "argument 3 (min_bonds) must be an integer")
        assertErrorMessage(self, "select oe_path_fp('C', 128, 0, '*')",
                           "argument 4 (max_bonds) must be an integer")
        assertErrorMessage(self, "select oe_path_fp('C', 128, -1, 5)",
                           "argument 3 (min_bonds) must be between")
        assertErrorMessage(self, "select oe_path_fp('C', 128, 4294967400, 5)",
                           "argument 3 (min_bonds) must be between")
        assertErrorMessage(self, "select oe_path_fp('C', 128, 0, -1)",
                           "argument 4 (max_bonds) must be between")
        assertErrorMessage(self, "select oe_path_fp('C', 128, 0, 4294967400)",
                           "argument 4 (max_bonds) must be between")
        assertErrorMessage(self, "select oe_path_fp('CCO', 16, 0)",
                           "argument 4 (max_bonds) is required if argument 3 is given");

    def test_smiles_live_database(self):
        self.assertEquals(
            L("select oe_path_fp(smiles, 16) from test_data where sid = 1"),
            (("0020", ),))
        self.assertEquals(
            L("select oe_path_fp(smiles, 16) from test_data where sid = 11"),
            ((None, ),))

    def test_num_bits_live_database(self):
        self.assertEquals(
            L("select oe_path_fp(smiles, 16) from test_data where sid = 1"),
            (("0020", ),))
        self.assertEquals(
            L("select oe_path_fp(smiles, 16) from test_data where sid = 11"),
            ((None, ),))
        self.assertEquals( # Don't allow 5 (less than 16)
            L("select oe_path_fp(smiles, sid) from test_data where sid = 5"),
            ((None, ),))
        self.assertEquals( # Don't allow 8 (still less than 16)
            L("select oe_path_fp(smiles, sid+3) from test_data where sid = 5"),
            ((None, ),))
        self.assertEquals( # Don't allow 20 (it's not a multiple of 8)
            L("select oe_path_fp(smiles, 15+sid) from test_data where sid = 5"),
            ((None, ),))
        self.assertEquals( # Allow 16
            L("select oe_path_fp(smiles, sid+11) from test_data where sid = 5"),
            (("fff1", ),))
        self.assertEquals( # huge
            L("select oe_path_fp(smiles, 4294967296+32+sid-5) from test_data where sid = 5"),
            ((None, ),))
        self.assertEquals( # negative number of bits aren't allowed
            L("select oe_path_fp(smiles, -128+sid-5) from test_data where sid = 5"),
            ((None, ),))
        self.assertEquals( # negative number of bits aren't allowed
            L("select oe_path_fp(smiles, sid-6) from test_data where sid = 5"),
            ((None, ),))

    def test_min_max_bonds_live_database(self):
        self.assertEquals(  # 0 to 5
            L("select oe_path_fp(smiles, 32, sid-5, sid) from test_data where sid = 5"),
            (("ddbffdf5", ),))
        self.assertEquals(  # 4 to 5; should give different results
            L("select oe_path_fp(smiles, 32, sid-1, sid) from test_data where sid = 5"),
            (("dd97e9b0", ),))
        self.assertEquals(  # 0 to 3; should give different results
            L("select oe_path_fp(smiles, 32, sid-5, sid-2) from test_data where sid = 5"),
            (("9dbf7c75", ),))
        self.assertEquals(  # just the atoms, at 0 to 0
            L("select oe_path_fp(smiles, 32, sid-5, sid-5) from test_data where sid = 5"),
            (("04004004", ),))
        self.assertEquals(  # don't allow min_bonds > max_bonds
            L("select oe_path_fp(smiles, 32, sid+1, sid) from test_data where sid = 1"),
            ((None, ),))
        self.assertEquals(  # error: negative min_bonds
            L("select oe_path_fp(smiles, 16, -sid, 5) from test_data where sid = 2"),
            ((None, ),))
        self.assertEquals(  # error: negative max_bonds
            L("select oe_path_fp(smiles, 16, 0, -sid) from test_data where sid = 2"),
            ((None, ),))
        self.assertEquals( # error: too large (shouldn't go negative)
            L("select oe_path_fp(smiles, 16, 2147483647+sid-5, 5) from test_data where sid = 5"),
            ((None, ),))
        self.assertEquals( # error: too large (shouldn't roll over)
            L("select oe_path_fp(smiles, 16, 4294967296+sid-5, 5) from test_data where sid = 5"),
            ((None, ),))

    def test_types(self):
        self.assertEquals(E("select oe_path_fp('c1ccccc1N', 32, 0, 5, 191, 3) = "
                            "       oe_path_fp('c1ccccc1N', 32, 0, 5)"), (1,))
        self.assertEquals(E("select oe_path_fp('c1ccccc1N', 32, 0, 5, 191) = "
                            "       oe_path_fp('c1ccccc1N', 32, 0, 5)"), (1,))
        self.assertEquals(E("select oe_path_fp('c1ccccc1N', 32, 0, 5, 191, 4) = "
                            "       oe_path_fp('c1ccccc1N', 32, 0, 5)"), (0,))
        self.assertEquals(E("select oe_path_fp('c1ccccc1N', 32, 0, 5, 190, 4) = "
                            "       oe_path_fp('c1ccccc1N', 32, 0, 5)"), (0,))
        self.assertEquals(E("select oe_path_fp('c1ccccc1N', 32, 0, 5, 190, 4)"),
                          ("bf7cbf75",))
        self.assertEquals(E("select oe_path_fp('c1ccccc1N', 32, 0, 5, 190, 5)"),
                          ("64fe2f37",))
        self.assertEquals(E("select oe_path_fp('c1ccccc1N', 32, 0, 5, 191, 4)"),
                          ("6fdeffb7",))

        assertErrorMessage(self, "select oe_path_fp('C', 32, 0, 5, -1, 4)",
                           "5 (atom_type) must be between")
        assertErrorMessage(self, "select oe_path_fp('C', 32, 0, 5, 2147483648, 4)",
                           "5 (atom_type) must be between")
        assertErrorMessage(self, "select oe_path_fp('C', 32, 0, 5, 190, -1)",
                           "6 (bond_type) must be between")
        assertErrorMessage(self, "select oe_path_fp('C', 32, 0, 5, 190, 2147483648)",
                           "6 (bond_type) must be between")
        
        
    def test_good_types_live_database(self):
        self.assertEquals(L("select oe_path_fp('c1ccccc1N', 32, 0, 5, 190, sid-1) "
                            "from test_data where sid = 5")[0],
                          ("bf7cbf75",))
        self.assertEquals(L("select oe_path_fp('c1ccccc1N', 32, 0, 5, sid, 4) "
                            "from test_data where sid = 5")[0],
                          ("0680e9a4",))

        self.assertEquals(L("select oe_path_fp('c1ccccc1N', 32, 0, 5, -sid, 4) "
                            "from test_data where sid = 5")[0], (None,))
        self.assertEquals(L("select oe_path_fp('c1ccccc1N', 32, 0, 5, 190, -sid) "
                            "from test_data where sid = 5")[0], (None,))
    
    
class TestFPValid(unittest.TestCase):
    def test_fp_valid(self):
        self.assertEquals(E("select fp_valid('0123456789abcdefABCDEF')"), (1,))
        self.assertEquals(E("select fp_valid('0')"), (1,))
        self.assertEquals(E("select fp_valid('00')"), (1,))
        self.assertEquals(E("select fp_valid('')"), (1,))
    def test_fp_invalid(self):
        for i in range(256):
            c = chr(i)
            if c in "0123456789abcdefABCDEF":
                continue
            self.assertEquals(E("select fp_valid(%s)", c), (0,))

class TestFPContains(unittest.TestCase):
    def test_fp_contains(self):
        self.assertEquals(E("select fp_contains('', '')"), (1,))
        self.assertEquals(E("select fp_contains('1', '1')"), (1,))
        self.assertEquals(E("select fp_contains('1', '2')"), (0,))
        self.assertEquals(E("select fp_contains('1', '3')"), (0,))
        self.assertEquals(E("select fp_contains('3', '1')"), (1,))

    def test_all_hex_values(self):
        values = [(c, int(c, 16)) for c in "0123456789ABCDEFabcdef"]
        for c1, i1 in values:
            for c2, i2 in values:
                if i1 & i2 == i2:
                    self.assertEquals(E("select fp_contains(%s, %s)", (c1, c2)), (1,))
                else:
                    self.assertEquals(E("select fp_contains(%s, %s)", (c1, c2)), (0,))

    def test_bad_args(self):
        self.assertRaises(MySQLdb.Error, E, "select fp_contains()")
        self.assertRaises(MySQLdb.Error, E, "select fp_contains('A')")
        self.assertRaises(MySQLdb.Error, E, "select fp_contains('A', 'B', 'C')")
        self.assertRaises(MySQLdb.Error, E, "select fp_contains('A', 1)")
        self.assertRaises(MySQLdb.Error, E, "select fp_contains(1, 'A')")

    def test_length_mismatch(self):
        self.assertEquals(E("select fp_contains('A', 'AB')"), (None,))
        self.assertEquals(E("select fp_contains('AB', 'A')"), (None,))

    def test_upper_lower_cases(self):
        self.assertEquals(E("select fp_contains('a', 'A')"), (1,))
        self.assertEquals(E("select fp_contains('B', 'b')"), (1,))
        self.assertEquals(E("select fp_contains('fa', '0A')"), (1,))
        self.assertEquals(E("select fp_contains('B6', 'b2')"), (1,))

    # Removed because this takes a long time
    # Does all 256*256 possible combinations
    if 0:
        def test_fp_contains_single_byte(self):
            sql = "select fp_contains(%s, %s)"
            valid_hex = "0123456789abcdefABCDEF"
            fields = "".join(chr(i) for i in range(256))
            for c1 in fields:
                if c1 not in valid_hex:
                    for c2 in fields:
                        self.assertEquals(E(sql, (c1, c2)), (None,))
                else:
                    i1 = int(c1, 16)
                    for c2 in fields:
                        if c2 not in valid_hex:
                            self.assertEquals(E(sql, (c1, c2)), (None,))
                        else:
                            i2 = int(c2, 16)
                            expected = (((i1 & i2) == i2),)
                            self.assertEquals(E(sql, (c1, c2)), expected)

# Fingerprints generated by OpenBabel. See
# http://www.dalkescientific.com/writings/diary/archive/2008/06/27/generating_fingerprints_with_openbabel.html

fp1 = ("00054062 81009600 00100102 81010700 200c0000 00202850 "
       "03102000 28000404 10200882 001849e4 483c0024 50039002 "
       "1402801a 01b90000 00020540 010c100a 22a4c000 82000300 "
       "2ac02018 00002201 60102120 183c9630 2100a080 81500019 "
       "00041401 80008c00 00484810 90001080 040c8202 0006081b "
       "24020080 a2042610").replace(" ", "")
fp2 = ("00010000 01000800 00100100 00010700 000c0050 00204010 "
       "00000000 20000000 00000100 010008c0 00200000 50019000 "
       "1c00801a 01890000 00000000 02081008 00a20000 02000000 "
       "01c02010 00000201 00000020 18020a30 01000080 00500014 "
       "00001401 80002000 00088000 10001000 00000200 0006000e "
       "c0020000 02002200").replace(" ", "")

# The result from MySQL:
#   0.35323383084576998
# is slightly different than what I compute with Python:
#   0.35323383084577115
# Don't know why. This tests for "good enough"
class NearFloat:
    def __init__(self, value1, value2=None):
        if value2 is None:
            self.value = value1
        else:
            self.value = float(value1)/value2
    def __eq__(self, other):
        return (self.value - other) < 1E-12
T = NearFloat(71./201.)

class TestTanimoto(unittest.TestCase):
    def test_fp_tanimoto(self):
        self.assertEquals(E("select fp_tanimoto(%s, %s)", (fp1, fp2)), (T,))
        self.assertEquals(E("select fp_tanimoto(%s, %s)", (fp1.upper(), fp2)), (T,))
        self.assertEquals(E("select fp_tanimoto(%s, %s)", (fp1.upper(), fp2.upper())), (T,))
        self.assertEquals(E("select fp_tanimoto(%s, %s)", (fp1, fp2.upper())), (T,))

    def test_zero(self):
        self.assertEquals(E("select fp_tanimoto(%s, %s)", ("0f", "F0")), (0.0,))

    def test_divide_by_zero(self):
        self.assertEquals(E("select fp_tanimoto(%s, %s)", ("", "")), (0.0,))
        self.assertEquals(E("select fp_tanimoto(%s, %s)", ("0", "0")), (0.0,))
        self.assertEquals(E("select fp_tanimoto(%s, %s)", ("00", "00")), (0.0,))


class TestLiveDatabase(unittest.TestCase):
    def test_valid_smiles(self):
        # This should find all of the valid SMILES and ignore invalid and NULL SMILES
        self.assertEquals(
            L("select sid from test_data where oe_valid_smiles(smiles) = 1 order by sid"),
            ((1,), (2,), (3,), (4,), (5,), (6,), (7,), (20,), (21,)))

    def test_cansmiles(self):
        # These are all valid
        self.assertEquals(
            # the Oc1ccccc1N result will change if OpenEye changes their canonicalization algorithm
            L("select oe_cansmiles(smiles) from test_data where flags = 0 order by sid"),
            (("C",), ("C",), ("CC",), ("CCC",), ("c1ccc(c(c1)N)O",), ("[NH4+]",), ("[U]",))  )

        # These are valid but not in canonical form
        self.assertEquals(
            L("select oe_cansmiles(smiles) from test_data where flags = 2 order by sid"),
            (("c1ccccc1",), ("c1ccccc1",)) )

        # These contain invalid SMILES. MySQL returns NULL for an error row and
        # *every* row after the error row.
        result = L("select oe_cansmiles(smiles) from test_data")
        self.assertEquals(result[-1], (None,))
        result = L("select oe_cansmiles(smiles) from test_data order by sid")
        self.assertEquals(result[-1], (None,))
        result = L("select oe_cansmiles(smiles) from test_data order by sid desc")
        self.assertEquals(result[-1], (None,))
        result = L("select oe_cansmiles(smiles) from test_data where sid = 10")  # illegal SMILES
        self.assertEquals(result[-1], (None,))
        result = L("select oe_cansmiles(smiles) from test_data where flags = 4")  # NULL smiles
        self.assertEquals(result[-1], (None,))

    def test_aromaticity_perception(self):
        # Ensure I'm perceiving aromaticity
        self.assertEquals(E("select oe_cansmiles('C1=CC=CC=C1')"), ("c1ccccc1",))

    def test_isosmiles(self):
        # These are all valid
        self.assertEquals(
            L("select oe_isosmiles(smiles) from test_data where flags = 0 order by sid"),
            (("C",), ("C",), ("CC",), ("CCC",), ("c1ccc(c(c1)N)O",), ("[NH4+]",), ("[238U]",)) )

        # These are valid but not in iso/canonical form
        self.assertEquals(
            L("select oe_isosmiles(smiles) from test_data where flags = 2 order by sid"),
            (("c1ccccc1",), ("c1ccccc1",)) )

        # These contain invalid SMILES. MySQL returns NULL for an error row and
        # *every* row after the error row.
        result = L("select oe_isosmiles(smiles) from test_data")
        self.assertEquals(result[-1], (None,))
        result = L("select oe_isosmiles(smiles) from test_data order by sid")
        self.assertEquals(result[-1], (None,))
        result = L("select oe_isosmiles(smiles) from test_data order by sid desc")
        self.assertEquals(result[-1], (None,))
        result = L("select oe_isosmiles(smiles) from test_data where sid = 10")  # illegal SMILES
        self.assertEquals(result[-1], (None,))
        result = L("select oe_isosmiles(smiles) from test_data where flags = 4")  # NULL smiles
        self.assertEquals(result[-1], (None,))

    def test_keksmiles(self):
        # These are all valid. Note that I defined the Kekule form to *include* isomeric terms
        self.assertEquals(
            L("select oe_keksmiles(smiles) from test_data where flags = 0 order by sid"),
            (("C",), ("C",), ("CC",), ("CCC",), ("C=1C=CC(=C(C1)N)O",), ("[NH4+]",), ("[238U]",)) )

        # These are valid but not in canonical form
        self.assertEquals(
            L("select oe_keksmiles(smiles) from test_data where flags = 2 order by sid"),
            (("C=1C=CC=CC1",), ("C1=CC=CC=C1",)) )

        # These contain invalid SMILES. MySQL returns NULL for an error row and
        # *every* row after the error row.
        result = L("select oe_keksmiles(smiles) from test_data")
        self.assertEquals(result[-1], (None,))
        result = L("select oe_keksmiles(smiles) from test_data order by sid")
        self.assertEquals(result[-1], (None,))
        result = L("select oe_keksmiles(smiles) from test_data order by sid desc")
        self.assertEquals(result[-1], (None,))
        result = L("select oe_keksmiles(smiles) from test_data where sid = 10")  # illegal SMILES
        self.assertEquals(result[-1], (None,))
        result = L("select oe_keksmiles(smiles) from test_data where flags = 4")  # NULL smiles
        self.assertEquals(result[-1], (None,))

    def test_oe_matches(self):
        # constant SMILES, variable SMARTS
        result = L("select oe_matches('CN', smarts) from test_data where flags = 0 order by sid")
        self.assertEquals(result, ((1,), (0,), (1,), (0,), (0,), (1,), (0,)) )

        # with an error in the SMILES
        self.assertRaises(MySQLdb.Error, L,
                          "select oe_matches('CQN', smarts) from test_data where flags = 0")
        # with an error in the SMARTS
        result = L("select oe_matches('CN', smarts) from test_data where flags = 2")
        self.assertEquals(result, ((None,), (None,)) )


        # variable SMILES, constant SMARTS
        result = L("select oe_matches(smiles, '[C,U]') from test_data where flags = 0 order by sid")
        self.assertEquals(result, ((1,), (1,), (1,), (1,), (0,), (0,), (1,)) )

        # with an error in the SMILES
        result = L("select oe_matches(smiles, smarts) from test_data where flags = 1")
        self.assertEquals(result, ((None,), (None,)) )

        # with an error in the SMARTS
        self.assertRaises(MySQLdb.Error, L, "select oe_matches(smiles, '[C,U') from test_data")


        # variable SMILES and SMARTS
        result = L("select oe_matches(smiles, smarts) from test_data where flags = 0 order by sid")
        self.assertEquals(result, ((1,), (0,), (1,), (1,), (1,), (1,), (1,)) )

        # with an error in the SMILES
        result = L("""select oe_matches(smiles, smarts) from test_data
                         where sid = 1 or sid = 10 order by sid""")
        self.assertEquals(result[-1], (None,))

        # with an error in the SMARTS
        result = L("""select oe_matches(smiles, smarts) from test_data
                         where sid = 1 or sid = 20 order by sid""")
        # with an error in both
        result = L("select oe_matches(smiles, smarts) from test_data where flags = 3")
        self.assertEquals(result, ((None,), (None,)) )

        # the NULL cases
        result = L("select oe_matches('C', smarts) from test_data where flags = 4")
        self.assertEquals(result, ((None,), ))
        result = L("select oe_matches(smiles, 'C') from test_data where flags = 4")
        self.assertEquals(result, ((None,), ))

    def test_oe_count_matches(self):
        # constant SMILES, variable SMARTS
        result = L("select oe_count_matches('CCN', smarts) from test_data where flags = 0 order by sid")
        self.assertEquals(result, ((2,), (2,), (2,), (2,), (0,), (1,), (0,)) )

        # with an error in the SMILES
        self.assertRaises(MySQLdb.Error, L,
                          "select oe_count_matches('CQN', smarts) from test_data where flags = 0")
        # with an error in the SMARTS
        result = L("select oe_count_matches('CN', smarts) from test_data where flags = 2")
        self.assertEquals(result, ((None,), (None,)) )


        # variable SMILES, constant SMARTS
        result = L("select oe_count_matches(smiles, '[C,U]') from test_data where flags = 0 order by sid")
        self.assertEquals(result, ((1,), (1,), (2,), (3,), (0,), (0,), (1,)) )
        result = L("select oe_count_matches(smiles, 'C[C,U]') from test_data where flags = 0 order by sid")
        self.assertEquals(result, ((0,), (0,), (2,), (4,), (0,), (0,), (0,)) )

        # with an error in the SMILES
        result = L("select oe_count_matches(smiles, smarts) from test_data where flags = 1")
        self.assertEquals(result, ((None,), (None,)) )

        # with an error in the SMARTS
        self.assertRaises(MySQLdb.Error, L, "select oe_count_matches(smiles, '[C,U') from test_data")


        # variable SMILES and SMARTS
        result = L("select oe_count_matches(smiles, smarts) from test_data where flags = 0 order by sid")
        self.assertEquals(result, ((1,), (0,), (2,), (4,), (6,), (1,), (1,)) )

        # with an error in the SMILES
        result = L("""select oe_count_matches(smiles, smarts) from test_data
                         where sid = 1 or sid = 10 order by sid""")
        self.assertEquals(result[-1], (None,))

        # with an error in the SMARTS
        result = L("""select oe_count_matches(smiles, smarts) from test_data
                         where sid = 1 or sid = 20 order by sid""")
        # with an error in both
        result = L("select oe_count_matches(smiles, smarts) from test_data where flags = 3")
        self.assertEquals(result, ((None,), (None,)) )

        # the NULL cases
        result = L("select oe_count_matches('C', smarts) from test_data where flags = 4")
        self.assertEquals(result, ((None,), ))
        result = L("select oe_count_matches(smiles, 'C') from test_data where flags = 4")
        self.assertEquals(result, ((None,), ))


    def test_oe_count_umatches(self):
        # constant SMILES, variable SMARTS
        result = L("select oe_count_umatches('CCN', smarts) from test_data where flags = 0 order by sid")
        self.assertEquals(result, ((2,), (1,), (2,), (1,), (0,), (1,), (0,)) )

        # with an error in the SMILES
        self.assertRaises(MySQLdb.Error, L,
                          "select oe_count_umatches('CQN', smarts) from test_data where flags = 0")
        # with an error in the SMARTS
        result = L("select oe_count_umatches('CN', smarts) from test_data where flags = 2")
        self.assertEquals(result, ((None,), (None,)) )


        # variable SMILES, constant SMARTS
        result = L("select oe_count_umatches(smiles, '[C,U]') from test_data where flags = 0 order by sid")
        self.assertEquals(result, ((1,), (1,), (2,), (3,), (0,), (0,), (1,)) )
        result = L("select oe_count_umatches(smiles, 'C[C,U]') from test_data where flags = 0 order by sid")
        self.assertEquals(result, ((0,), (0,), (1,), (2,), (0,), (0,), (0,)) )

        # with an error in the SMILES
        result = L("select oe_count_umatches(smiles, smarts) from test_data where flags = 1")
        self.assertEquals(result, ((None,), (None,)) )

        # with an error in the SMARTS
        self.assertRaises(MySQLdb.Error, L, "select oe_count_umatches(smiles, '[C,U') from test_data")


        # variable SMILES and SMARTS
        result = L("select oe_count_umatches(smiles, smarts) from test_data where flags = 0 order by sid")
        self.assertEquals(result, ((1,), (0,), (2,), (2,), (6,), (1,), (1,)) )

        # with an error in the SMILES
        result = L("""select oe_count_umatches(smiles, smarts) from test_data
                         where sid = 1 or sid = 10 order by sid""")
        self.assertEquals(result[-1], (None,))

        # with an error in the SMARTS
        result = L("""select oe_count_umatches(smiles, smarts) from test_data
                         where sid = 1 or sid = 20 order by sid""")
        self.assertEquals(result[-1], (None,))

        # with an error in both
        result = L("select oe_count_umatches(smiles, smarts) from test_data where flags = 3")
        self.assertEquals(result, ((None,), (None,)) )

        # the NULL cases
        result = L("select oe_count_umatches('C', smarts) from test_data where flags = 4")
        self.assertEquals(result, ((None,), ))
        result = L("select oe_count_umatches(smiles, 'C') from test_data where flags = 4")
        self.assertEquals(result, ((None,), ))


    def test_lingosim(self):
        result = L("select sid, oe_lingosim(smiles, 'c1ccccc1O') from test_data "
                   "where sid <= 5 order by sid")
        expected = (
            (1, 0.0),
            (2, 0.0),
            (3, 0.0),
            (4, 0.0),
            (5, NearFloat(0.83333331346511841)),
            )
        self.assertEquals(result, expected)
        result = L("select sid, oe_lingosim('c1ccccc1O', smiles) from test_data "
                   "where sid <= 5 order by sid")
        self.assertEquals(result, expected)


#    def test_oe_maccs_fp(self):
#        result = L("select oe_maccs_fp(smiles) from test_data where sid = 5")
#        self.assertEquals(result, (("00000000000002000000800400001004454C1E83F1",),) )
#        result = L("select oe_maccs_fp(smiles) from test_data where smiles = 'C'")
#        self.assertEquals(len(result), 2)
#        self.assertEquals(result[0], result[1])
        

    def test_fp_valid(self):
        result = L("select sid from test_data where fp_valid(fp) = 0 order by sid")
        self.assertEquals(result, ((6,), (40,)))

        result = L("select fp_valid(fp) from test_data where sid < 4")
        self.assertEquals(result, ((1,), (1,), (1,)))

    def test_fp_contains(self):
        result = L("select sid from test_data where sid < 4 and fp_contains(fp, '000f') order by sid")
        self.assertEquals(result, ((2,), (3,)))
        result = L("select sid from test_data where (sid = 10 or sid = 11) and "
                   "fp_contains(fp, '040000') order by sid")
        self.assertEquals(result, ((10,), (11,)))
        result = L("select sid from test_data where (sid = 10 or sid = 11) and "
                   "fp_contains(fp, '060000') order by sid")
        self.assertEquals(result, ((11,),))
        result = L("select sid from test_data where (sid = 10 or sid = 11) and "
                   "fp_contains(fp, '080000') order by sid")
        self.assertEquals(result, ())


    def test_fp_tanimoto(self):
        result = L("select sid, fp_tanimoto(fp, '1a0B') from test_data where length(fp) = 4 order by sid")
        expected = ( (1, 0.0,), # sid 1
                     (2, NearFloat(1+1+0+3, 1+2+4+4)),
                     (3, NearFloat(6, 16)),
                     (7, 0.0,),
                     (20, 0.0),
                     (21, NearFloat(2, 4+2+1+3)),
                     (30, NearFloat(0+1+0+1, 3+3+2+3)),
                     (31, NearFloat(1+1+0+2, 3+4+2+3)),
                     )
        self.assertEquals(result, expected)
        # And check that it's reflexive
        result = L("select sid, fp_tanimoto('1a0B', fp) from test_data where length(fp) = 4 order by sid")
        self.assertEquals(result, expected)



def create_test_database():
    c = C()
    c.execute("create database " + TEST_DATABASE);
    c = C(TEST_DATABASE)
    c.execute("""
CREATE TABLE test_data (
  sid INTEGER NOT NULL PRIMARY KEY,
  smiles TEXT,
  smarts TEXT,
  flags INTEGER,
  fp TEXT);
""");
  
test_data = [
    # valid SMILES and SMARTS
    (1, "C", "C", 0, "0000"),
    (2, "C", "CC", 0, "12FF"),
    (3, "CC", "C", 0, "ffff"),
    (4, "CCC", "CC", 0, "88"),
    (5, "Nc1ccccc1O", "c", 0, "9F"),
    (6, "[NH4+]", "N", 0, "@"),  # bad fingerprint
    (7, "[238U]", "[U]", 0, "0000"), # isomeric

    # invalid SMILES/valid SMARTS
    (10, "Q", "C", 1, "343434"),
    (11, "C(", "[C,N]", 1, "373737"),

    # valid SMILES/invalid SMARTS
    (20, "C1=CC=CC=C1", "Q", 2, "E000"),
    (21, "C=1C=CC=CC1", "[C", 2, "F012"),

    # invalid SMILES and invalid SMARTS
    (30, "Q", "Q", 3, "cc31"),
    (31, "(", "[", 3, "dd33"),
    
    # NULL fields
    (40, None, None, 4, None),
    ]
for (x, s1, s2, flg, fp) in test_data:
    assert isinstance(x, int)
    assert isinstance(s1, str) or flg == 4
    assert isinstance(s2, str) or flg == 4
    assert isinstance(flg, int)
    assert isinstance(fp, str) or flg == 4

def populate_test_database():
    c = C(TEST_DATABASE)
    c.executemany(
        "INSERT INTO test_data (sid, smiles, smarts, flags, fp) VALUES (%s, %s, %s, %s, %s)",
        test_data)

def check_connection():
    try:
        result = E("select 1")
    except:
        traceback.print_exc()
        sys.exit("Could not connect to the MySQL database.\n"
                 "Is the server running and does this program have the right connection settings?")

    try:
        result = E("select oe_licensed()")
    except:
        traceback.print_exc()
        sys.exit("MySQL is running and I can connect to it, but "
                 "the cyclops_mysql extension does not seem to be installed")

    if result == 0:
        for name in ("oechem", "oegraphsim"):
            if E("select oe_licensed('" + name + "')") == 0:
                expired = E("select oe_license_date('" + name + "')")[0]
                sys.exit("The cyclops_mysql plugin is installed but the %s license expired on %s." % (name, expired))
        raise AssertionError("How did the license fail but not for the components?")
        

if __name__ == "__main__":
    import sys
    if "--create-database" in sys.argv[1:]:
        if len(sys.argv) > 2:
            raise SystemExit("no other arguments allowed with --create-database")
        create_test_database()
        populate_test_database()
        sys.exit()

    if "--no-database" in sys.argv[1:]:
        if len(sys.argv) > 2:
            raise SystemExit("no other arguments allowed with --no-database")
        del TestLiveDatabase
        del TestPathFP.test_good_types_live_database
        del TestPathFP.test_min_max_bonds_live_database
        del TestPathFP.test_num_bits_live_database
        del TestPathFP.test_smiles_live_database
        del sys.argv[1]
        
    # Check that I can connect to the database. This is a sanity check
    check_connection()

    unittest.main()
